# -*- coding: utf-8 -*-
# tests/utils.py


import os
from pathlib import Path

from jupyter_analysis_tools.utils import (
    appendToPATH,
    isWindows,
    makeNetworkdriveAbsolute,
    naturalKey,
    networkdriveMapping,
)

# output of 'net use' command on Windows
outNetUse = r"""Neue Verbindungen werden gespeichert.


Status       Lokal     Remote                    Netzwerk

-------------------------------------------------------------------------------
OK           G:        \\ALPHA\BETA              Microsoft Windows Network
OK           K:        \\GAM\MMA                 Microsoft Windows Network
OK           M:        \\user\drive\uname        Microsoft Windows Network
OK           T:        \\test\foldername         Microsoft Windows Network
OK                     \\psi\folder              Microsoft Windows Network
Der Befehl wurde erfolgreich ausgeführt.
"""

# sample output of 'mount' command on Linux
outMount = (
    "sysfs on /sys type sysfs (rw,nosuid,nodev,noexec,relatime)\n"
    "proc on /proc type proc (rw,nosuid,nodev,noexec,relatime)\n"
    "tmpfs on /run type tmpfs (rw,nosuid,nodev,noexec,relatime,size=13148680k,mode=755,inode64)\n"
    "//abc02.def.ault.de/X23/somename on /mnt/some (ugly) on type name type cifs "
    "(rw,nosuid,nodev,relatime,vers=3.0,cache=strict,upcall_target=app,username=dhdhfh,"
    "uid=1000,forceuid,gid=1000,forcegid,addr=10.0.1.2,file_mode=0660,dir_mode=0770,soft,nounix,"
    "mapposix,rsize=4194304,wsize=4194304,bsize=1048576,retrans=1,echo_interval=60,actimeo=1,"
    "closetimeo=1)\n"
    "udev on /dev type devtmpfs (rw,nosuid,relatime,size=65700820k,nr_inodes=16425205,mode=755,"
    "inode64)\n"
    "//xyz04.fgsd.asd.com/G2S/GH31 on /mnt/gh 12 type cifs (rw,nosuid,nodev,relatime,vers=3.0,"
    "cache=strict,upcall_target=app,username=dhdhfh,uid=1000,forceuid,gid=1000,forcegid,"
    "addr=10.6.1.5,file_mode=0660,dir_mode=0770,soft,nounix,mapposix,rsize=4194304,wsize=4194304,"
    "bsize=1048576,retrans=1,echo_interval=60,actimeo=1,closetimeo=1)\n"
    "devpts on /dev/pts type devpts (rw,nosuid,noexec,relatime,gid=5,mode=620,ptmxmode=000)"
)


def test_appendToPATH(capsys):
    # Setting up a PATH for testing first (platform dependent).
    testpath = "/usr/local/sbin:/usr/local/bin:/sbin:/usr/games:/usr/local/games:/snap/bin"
    if isWindows():
        testpath = "something else"
    os.environ["PATH"] = testpath
    assert os.environ["PATH"] == testpath

    if not isWindows():  # Linux edition
        appendToPATH("/tmp", ("one", "two"), verbose=True)
        captured = capsys.readouterr()
        assert (
            captured.out
            == """\
     /tmp/one [exists: False]
     /tmp/two [exists: False]
"""
        )
        assert os.environ["PATH"] == testpath + ":/tmp/one:/tmp/two"

    else:  # Windows edition
        appendToPATH(r"C:\Windows", ("one", "two"), verbose=True)
        captured = capsys.readouterr()
        assert (
            captured.out
            == """\
     C:\\Windows\\one [exists: False]
     C:\\Windows\\two [exists: False]
"""
        )
        assert os.environ["PATH"] == testpath + r";C:\Windows\one;C:\Windows\two"


def test_networkdriveMapping():
    if isWindows():
        map = networkdriveMapping(cmdOutput=outNetUse, resolveNames=False)
        assert map == {
            "G:": "\\\\ALPHA\\BETA",
            "K:": "\\\\GAM\\MMA",
            "M:": "\\\\user\\drive\\uname",
            "T:": "\\\\test\\foldername",
        }
    else:  # Linux or macOS
        map = networkdriveMapping(cmdOutput=outMount, resolveNames=False)
        assert map == {
            "/mnt/gh 12": "//xyz04.fgsd.asd.com/G2S/GH31",
            "/mnt/some (ugly) on type name": "//abc02.def.ault.de/X23/somename",
        }


def test_makeNetworkdriveAbsolute():
    if isWindows():
        filepath = Path(r"M:\some\folders\a file name.ext")
        newpath = makeNetworkdriveAbsolute(filepath, cmdOutput=outNetUse, resolveNames=False)
        assert filepath != newpath
        assert newpath == Path(r"\\user\drive\uname\some\folders\a file name.ext")
    else:  # Linux or macOS
        filepath = Path("/mnt/some (ugly) on type name/some/folders/a file name.ext")
        newpath = makeNetworkdriveAbsolute(filepath, cmdOutput=outMount, resolveNames=False)
        assert filepath != newpath
        assert newpath == Path("//abc02.def.ault.de/X23/somename/some/folders/a file name.ext")


def test_naturalKey():
    filelist = ["test2.ext", "test100.ext", "test1.ext", "test05.ext"]
    lstSorted = sorted(filelist, key=naturalKey)
    assert lstSorted == ["test1.ext", "test2.ext", "test05.ext", "test100.ext"]



# -*- coding: utf-8 -*-


import subprocess
from os.path import abspath, dirname, join

import toml

base_path = dirname(dirname(abspath(__file__)))
project_meta = toml.load(join(base_path, "pyproject.toml"))

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.extlinks",
    "sphinx.ext.ifconfig",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx.ext.inheritance_diagram",
    "sphinx.ext.graphviz",
    "myst_parser",
]
inheritance_edge_attrs = dict(color="gray")  # readable in darkmode too
autosummary_generate = True  # Turn on sphinx.ext.autosummary
templates_path = ["_templates"]
source_suffix = ".rst"
master_doc = "index"
project = "Jupyter Analysis Tools"
year = "2018-2025"
author = "Ingo Breßler and Brian R. Pauw"
copyright = "{0}, {1}".format(year, author)
version = "1.4.3"
release = version
commit_id = None
try:
    commit_id = (
        subprocess.check_output(["git", "rev-parse", "--short", "HEAD"]).strip().decode("ascii")
    )
except subprocess.CalledProcessError as e:
    print(e)


autodoc_mock_imports = [
    "ipykernel",
    "notebook",
    "pandas",
    "ipywidgets",
    "matplotlib",
    "scipy",
    "h5py",
    "pint",
    "sasmodels",
    "chempy",
    "graphviz",
    "mcsas3",
]

pygments_style = "trac"
extlinks = {
    "issue": (join(project_meta["project"]["urls"]["repository"], "issues", "%s"), "#%s"),
    "pr": (join(project_meta["project"]["urls"]["repository"], "pull", "%s"), "PR #%s"),
}
html_theme = "furo"

html_use_smartypants = True
html_last_updated_fmt = "%b %d, %Y"
if commit_id:
    html_last_updated_fmt += f" (git {commit_id})"
html_split_index = False

html_short_title = "%s-%s" % (project, version)

napoleon_use_ivar = True
napoleon_use_rtype = False
napoleon_use_param = False

linkcheck_ignore = [
    join(
        project_meta["project"]["urls"]["documentation"],
        project_meta["tool"]["coverage"]["report"]["path"],
    )
    + r".*",
    # attempted fix of '406 Client Error: Not Acceptable for url'
    # https://github.com/sphinx-doc/sphinx/issues/1331
    join(project_meta["project"]["urls"]["repository"], "commit", r"[0-9a-fA-F]+")
]
linkcheck_anchors_ignore_for_url = [
    r"https://pypi\.org/project/[^/]+",
]



# -*- coding: utf-8 -*-
# datalocations.py

import glob
import os
import shutil
import tempfile
from pathlib import Path

from .utils import indent, isList


def getWorkDir(workDir=None, skip=False):
    """Find a local work dir for temporary files, created during analysis.
    The default is *$HOME/data*."""
    if skip:  # stay in the current directory if desired
        return os.path.abspath(".")
    if not workDir or not len(workDir):
        workDir = Path.home() / "data"
    else:
        workDir = Path(workDir).resolve()
    if not workDir.is_dir():
        os.mkdir(workDir)
    print("Using '{}' as working directory.".format(workDir))
    return workDir


def prepareWorkDir(workDir, srcDir, useExisting=False):
    """Create a temporary working directory and copy
    the input data (series) to it if not already present."""
    # source dir has to exist
    if not os.path.isdir(srcDir):
        raise RuntimeError("Provided source directory '{}' not found!".format(srcDir))
    srcDir = os.path.realpath(srcDir)
    # no separate work dir requested?
    if os.path.samefile(workDir, os.getcwd()):
        print("Working in current directory '{}'.".format(os.getcwd()))
        return srcDir  # nothing to do
    prefix = os.path.basename(srcDir) + "_"
    if useExisting:  # use an existing work dir, avoid copying
        dirs = glob.glob(os.path.join(workDir, prefix + "*"))
        if len(dirs):
            return dirs[0]  # use the first match
        print("No existing work dir found, creating a new one.")
    # copy all data from src dir to a newly created work dir
    workDir = tempfile.mkdtemp(dir=workDir, prefix=prefix)
    print("Copying data to {}:".format(workDir))
    for dn in os.listdir(srcDir):
        srcPath = os.path.join(srcDir, dn)
        dstPath = os.path.join(workDir, dn)
        if os.path.isdir(srcPath):
            shutil.copytree(srcPath, dstPath)
            print(indent, dn)
        if os.path.isfile(srcPath):
            shutil.copy(srcPath, dstPath)
            print(indent, dn)
    print("Done preparing work dir.")
    return workDir


def printFileList(fnlst, numParts=2, limit=20):
    def printlst(lst):
        return [print(indent, fn) for fn in lst]

    def shorten(lst):
        return [os.path.join(*Path(fn).parts[-numParts:]) for fn in lst]

    if len(fnlst) > limit:
        printlst(shorten(fnlst[:3]))
        print(indent, "[...]")
        printlst(shorten(fnlst[-3:]))
    else:
        printlst(shorten(fnlst))


def getDataDirs(dataDir, noWorkDir=False, reuseWorkDir=True, workDir=None):
    """Create a local work dir with a copy of the input data and for storing the results.
    (Data might reside in synced folders which creates massive traffic once batch processing
    results get replaced repeately.)

    Parameters
    ----------
    noWorkDir: bool
        False: Copy input data to a new working dir (default),
        True: otherwise, use data where it is.
    reuseWorkDir: bool
        False: Create a new working dir each time,
        True: reuse the work dir if it exists already (default).

    Returns
    -------
    A list of absolute directory paths.
    """
    basedir = getWorkDir(workDir=workDir, skip=noWorkDir)
    workDir = prepareWorkDir(basedir, dataDir, useExisting=reuseWorkDir)
    print("Entering '{}':".format(workDir))
    dirs = sorted([dn for dn in Path(workDir).iterdir() if dn.is_dir()])
    dirs.append(Path(workDir))
    # [print(os.path.join(*dn.parts[-2:])) for dn in dirs]
    printFileList(dirs, numParts=1)
    return dirs


def getDataFiles(dataDirs, include=None, exclude=None):
    """Return absolute file paths from given directories."""

    def getFiles(dn, include=None):
        if not include:
            include = "*"
        if not isList(include):
            include = (include,)
        return [path for inc in include for path in glob.glob(os.path.join(dn, inc))]

    if not exclude:
        exclude = ()
    if not isList(exclude):
        exclude = (exclude,)
    if not isList(dataDirs):
        dataDirs = (dataDirs,)

    files = [
        fn
        for dn in dataDirs
        for fn in getFiles(dn, include)
        if not any([(ex in fn) for ex in exclude])
    ]
    print("{} files to be analyzed in subdirectories.".format(len(files)))
    return sorted(files)



# -*- coding: utf-8 -*-
# widgets.py

import os

import ipywidgets as ui


def showBoolStatus(value, description, invertcolor=False):
    """Pretty prints the status of a boolean variable *value* along with
    the provided description in a green color for True and in a red color
    for False values. *invertcolor* allows to flip the color assignment."""
    from IPython.display import HTML, display

    statuscolor = "darkgreen"
    if value ^ invertcolor:
        statuscolor = "darkred"
    descr = description[0].lower() + description[1:-1]
    if description[-1].isalnum():
        descr += description[-1]
    statustext = '<h4 style="color: {};">Yes, {}!</h4>'.format(statuscolor, descr)
    if value:
        statustext = '<h4 style="color: {};">Do not {}!</h4>'.format(statuscolor, descr)
    display(
        HTML(
            f'<div style="border-style: solid; border-color: {statuscolor};'
            " border-width: 1px; padding: 0em 1em .5em 1em; margin: 1em 0em;"
            f' width: {len(descr) * 0.75}em;">'
            + statustext
            + "</div>"
        )
    )


class PathSelector:
    def __init__(self, start_dir, select_file=True):
        self.file = None
        self.select_file = select_file
        self.cwd = start_dir
        self.select = ui.SelectMultiple(options=["init"], value=(), rows=10, description="")
        self.accord = ui.Accordion(children=[self.select])

        self.accord.selected_index = None  # Start closed (showing path only)
        self.refresh(self.cwd)
        self.select.observe(self.on_update, "value")

    def on_update(self, change):
        if len(change["new"]) > 0:
            self.refresh(change["new"][0])

    def refresh(self, item):
        path = os.path.abspath(os.path.join(self.cwd, item))

        if os.path.isfile(path):
            if self.select_file:
                self.accord.set_title(0, path)
                self.file = path
                self.accord.selected_index = None
            else:
                self.select.value = ()

        else:  # os.path.isdir(path)
            self.file = None
            self.cwd = path

            # Build list of files and dirs
            keys = ["[..]"]
            for item in os.listdir(path):
                if item[0] == ".":
                    continue
                elif os.path.isdir(os.path.join(path, item)):
                    keys.append("[" + item + "]")
                else:
                    keys.append(item)

            # Sort and create list of output values
            keys.sort(key=str.lower)
            vals = []
            for k in keys:
                if k[0] == "[":
                    vals.append(k[1:-1])  # strip off brackets
                else:
                    vals.append(k)

            # Update widget
            self.accord.set_title(0, path)
            self.select.options = list(zip(keys, vals))
            with self.select.hold_trait_notifications():
                self.select.value = ()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# binning.py

"""
Overview
========
1D rebinning.
Should take input file, read, rebin and write.
Rebins to log bins.
"""

# __author__ = "Brian R. Pauw"
# __contact__ = "brian@stack.nl"
# __license__ = "GPLv3+"
# __date__ = "2015/01/09"
# __status__ = "beta"

import argparse
import itertools
import os
import sys

import numpy as np
import pandas
from numpy import argsort, log10, reshape, shape, size, sqrt, zeros


def argparser():
    parser = argparse.ArgumentParser(
        description="""
            Re-binning function, reads three-column ASCII input files,
            and outputs re-binned three-column ASCII files"""
    )
    # binning options
    parser.add_argument("-n", "--numBins", type=int, default=50, help="Number of bins to use")
    parser.add_argument(
        "-q",
        "--qMin",
        type=float,
        default=0.0,
        help="Minimum Q to clip from original data",
    )
    parser.add_argument(
        "-Q",
        "--qMax",
        type=float,
        default=np.inf,
        help="Minimum Q to clip from original data",
    )
    parser.add_argument(
        "-e",
        "--minE",
        type=float,
        default=0.01,
        help="Minimum error is at least this times intensity value.",
    )
    parser.add_argument(
        "-s",
        "--scaling",
        type=str,
        action="store",
        default="logarithmic",
        help="q-axis scaling for binning, can be linear or logarithmic",
    )
    # csv / datafile options
    parser.add_argument(
        "-d",
        "--delimiter",
        type=str,
        action="store",
        default=",",
        help="Delimiter in original file. '\\t' is tab. (with quotes)",
    )
    parser.add_argument(
        "-H",
        "--headerLines",
        type=int,
        default=0,
        help="Number of header lines to skip",
    )
    parser.add_argument(
        "-D",
        "--outputDelimiter",
        type=str,
        action="store",
        default=None,
        help="Delimiter in final file (defaults to input delimiter)",
    )
    parser.add_argument(
        "-c",
        "--cleanEmpty",
        action="store_true",
        default=True,
        help="Removes empty bins before writing",
    )
    parser.add_argument(
        "-i",
        "--iScale",
        type=float,
        default=1.0,
        help="Intensity (and error) scaled by this factor on output.",
    )
    # program options
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Be verbose about the steps",
    )
    parser.add_argument(
        "-t",
        "--test",
        action="store_true",
        help="Do not save output files, test run only",
    )
    parser.add_argument(
        "-N",
        "--noBin",
        action="store_true",
        help="Do not bin, just input -> output (for translation and scaling)",
    )
    parser.add_argument(
        "fnames",
        nargs="*",
        metavar="FILENAME",
        action="store",
        help="One or more data files to rebin",
    )
    # show help if no files were provided, no arguments at all
    args = parser.parse_args()
    if len(args.fnames):
        return args
    parser.print_help(sys.stderr)
    sys.exit(1)


class reBin(object):
    """all kinds of binning-related functions"""

    # set defaults for file reading:
    pandasArgs = {
        "skipinitialspace": True,
        "skip_blank_lines": True,
        "engine": "python",
        "header": None,
    }
    # set defaults for kwargs, in case this is not called from command line:
    reBinArgs = {
        "delimiter": ";",
        "outputDelimiter": ";",
        "headerLines": 0,
        "fnames": "",
        "verbose": False,
        "qMin": -np.inf,
        "qMax": np.inf,
        "numBins": 100,
        "scaling": "logarithmic",
        "cleanEmpty": False,
        "minE": 0.01,
        "noBin": False,
    }

    def __init__(self, **kwargs):
        # process defaults:
        for kw in self.reBinArgs:
            setattr(self, kw, self.reBinArgs[kw])
        # process kwargs:
        if "verbose" in kwargs:
            self.verbose = kwargs.pop("verbose")
        for kw in kwargs:
            if self.verbose:
                print("Processing input argument {}: {}".format(kw, kwargs[kw]))
            setattr(self, kw, kwargs[kw])

        # process delimiter options
        # decode no longer necessary in python 3
        if sys.version_info <= (3, 0):
            self.delimiter = self.delimiter.decode("string-escape")
        if self.outputDelimiter is None:
            self.outputDelimiter = self.delimiter
        else:
            if sys.version_info <= (3, 0):
                self.outputDelimiter = self.outputDelimiter.decode("string-escape")

        self.pandasArgs.update({"delimiter": self.delimiter, "skiprows": self.headerLines})
        # process files individually:
        for filename in self.fnames:
            self.readFile(filename)
            self.validate()
            self.defineBinEdges()
            self.binning1D()
            if self.cleanEmpty:
                # removes bins with no intensity or error
                self.cleanup()

            if not self.test:
                # generate output file name
                ofname = self.outputFilename(filename)
                # write binned data to file name
                self.writeFile(ofname)

    def cleanup(self):
        # removes unwanted bin values
        # cannot use lists, because:
        # http://unspecified.wordpress.com/2009/02/12/thou-shalt-not-modify-a-list-during-iteration
        validi = True ^ np.isnan(self.IBin)
        validi[np.argwhere(self.binMask > 0)] = False
        self.QBin = self.QBin[validi]
        self.IBin = self.IBin[validi]
        self.EBin = self.EBin[validi]
        self.QEBin = self.QEBin[validi]
        if self.verbose:
            print("valid bins: {} of {}".format(validi.sum(), len(validi)))

    def outputFilename(self, filename):
        """returns an output filename based on the input filename"""
        of = filename.strip()
        # split at extension
        ob, oe = of.rsplit(".", 1)
        # add rebin tag and reassemble
        ofname = "{}_reBin.{}".format(ob, oe)
        if self.verbose:
            print("output filename: {}".format(ofname))
        return ofname

    def readFile(self, filename):
        if self.verbose:
            print("reading file: {} with settings: {}".format(filename, self.pandasArgs))
        dval = pandas.read_csv(filename, **self.pandasArgs).values
        assert isinstance(dval, np.ndarray)  # no problems reading?
        assert size(dval, axis=1) >= 3  # Q, I and E can be extracted
        if self.verbose:
            print("data read: {}".format(dval))
        self.Q = np.float32(dval[:, 0])
        self.I = np.float32(dval[:, 1])
        self.E = np.maximum(self.minE * self.I, np.float32(dval[:, 2]))
        numChanged = (self.minE * self.I > dval[:, 2]).sum()
        if self.verbose:
            print(
                "Minimum uncertainty set for {} out of {} ({} %) datapoints".format(
                    numChanged, size(self.Q), 100.0 * numChanged / size(self.Q)
                )
            )

    # writer modified from imp2/modules/Write1D
    def writeFile(self, ofname, hstrs=None, append=False):
        sep = self.outputDelimiter
        # scale if necessary
        iterData = itertools.zip_longest(
            self.QBin,
            self.IBin * float(self.iScale),
            self.EBin * float(self.iScale),
        )

        def writeLine(filename, line=None, append=True):
            if append:
                openarg = "a"
            else:
                openarg = "w"
            with open(filename, openarg) as fh:
                if isinstance(line, str):
                    fh.write(line)
                else:
                    # iterable object containing multiple lines
                    fh.writelines(line)

        # truncate file if exists (i.e. discard)
        if os.path.exists(ofname) and (not append):
            os.remove(ofname)

        # write header and data:
        if hstrs is not None:
            writeLine(ofname, hstrs)

        # store in file
        moreData = True
        while moreData:
            try:
                # generate formatted datastring containing column data
                wstr = sep.join(["{}".format(k) for k in next(iterData)]) + "\n"
            except StopIteration:
                # end of data reached
                moreData = False
                break
            writeLine(ofname, wstr)

    def validate(self):
        """Applies limits to the data"""
        mask = zeros(shape(self.Q), dtype="bool")
        # appy integration limits:
        iind = np.array(((self.Q < self.qMin) + (self.Q > self.qMax)), dtype=bool)
        mask[iind] = True

        # define binning limits
        (qmin, qmax) = (
            np.abs(self.Q[True ^ mask]).min(),
            np.abs(self.Q[True ^ mask]).max(),
        )
        self.iqMin = np.maximum(qmin, self.qMin)
        self.iqMax = np.minimum(qmax, self.qMax)
        self.Q = self.Q[True ^ mask]
        self.I = self.I[True ^ mask]
        self.E = self.E[True ^ mask]
        if self.verbose:
            print(
                "data Q-range: {}, integration Q-range: {}, masked: {} of {} ({}%)".format(
                    (self.Q.min(), self.Q.max()),
                    (self.iqMin, self.iqMax),
                    mask.sum(),
                    self.Q.size,
                    mask.sum() / self.Q.size,
                )
            )

    def defineBinEdges(self):
        """defines binning edges"""
        # define bin edges
        if self.scaling.lower() in ("linear", "lin"):
            qEdges = np.linspace(self.iqMin, self.iqMax, self.numBins + 1)
        else:
            qEdges = np.logspace(log10(self.iqMin), log10(self.iqMax), self.numBins + 1)
        self.qEdges = qEdges
        if self.verbose:
            print("Bin edges used: {}".format(self.qEdges))

    def binning1D(self, qError=None):
        """An unweighted binning routine.
        imp-version of binning, taking q-bin edges in which binning takes place,
        and calculates the mean q uncertainty in the bin as well from the relative
        Q uncertainties provided.

        The intensities are sorted across bins of equal size. If provided error
        is empty, the standard deviation of the intensities in the bins is
        computed.
        """
        # no binning requested, just input -> output
        if self.noBin:
            self.QBin = self.Q.copy()
            self.IBin = self.I.copy()
            self.EBin = self.E.copy()
            self.QEBin = np.zeros(np.shape(self.I))
            return

        # set values:
        q = self.Q.copy()
        intensity = self.I.copy()
        error = self.E.copy()
        numBins = self.numBins
        qEdges = self.qEdges

        # flatten q, intensity and error
        q = reshape(q, size(q))
        intensity = reshape(intensity, size(intensity))

        # sort q, let intensity and error follow sort
        sortInd = argsort(q, axis=None)
        q = q[sortInd]
        intensity = intensity[sortInd]

        # initialise storage:
        numBins = len(qEdges) - 1
        ibin = zeros(numBins)
        qbin = zeros(numBins)
        sdbin = zeros(numBins)
        sebin = zeros(numBins)
        qebin = zeros(numBins)
        binMask = zeros(numBins)  # set one for masked bin values
        if error is not None:
            error = reshape(error, size(error))
            error = error[sortInd]
        if qError is not None:
            qError = reshape(qError, size(qError))
            qError = qError[sortInd]

        # now we can fill the bins
        for bini in range(numBins):
            # limit ourselves to only the bits we're interested in:
            limMask = (q >= qEdges[bini]) & (q <= qEdges[bini + 1])

            iToBin = intensity[limMask]
            # sum the intensities in one bin and normalize by number of pixels
            if limMask.sum() == 0:
                # no pixels in bin
                (ibin[bini], sebin[bini], qebin[bini], qbin[bini]) = (
                    None,
                    None,
                    None,
                    None,
                )
                binMask[bini] = 1
                continue

            elif limMask.sum() == 1:
                ibin[bini] = iToBin.mean()
                qbin[bini] = q[limMask].mean()
                if error is not None:
                    sebin[bini] = error[limMask]
                if qError is not None:
                    qebin[bini] = qError[limMask]

            else:
                ibin[bini] = iToBin.mean()
                qbin[bini] = q[limMask].mean()
                if error is not None:
                    sebin[bini] = np.sqrt((error[limMask] ** 2).sum()) / limMask.sum()
                # now we deal with the Errors:
                # calculate the standard deviation of the intensity in the bin
                # according to the definition of sample-standard deviation
                sdbin = iToBin.std(ddof=1)
                # what we want is to have the "standard error of the mean":
                sdbin = sdbin / sqrt(1.0 * np.size(iToBin))
                # maximum between standard error and Poisson statistics
                sebin[bini] = np.maximum(sebin[bini], sdbin)
                # qebin is the mean error of the q-values in the bin, should
                # probably be superseded by the bin width
                qe = 0.0
                if qError is not None:
                    qe = np.sqrt((qError[limMask] ** 2).sum())
                # SSTD of q in bin:
                qs = np.std(q[limMask], ddof=1)  # sample standard deviation
                qebin[bini] = np.maximum(qe, qs)

        self.QBin = qbin.copy()
        self.IBin = ibin.copy()
        self.EBin = sebin.copy()
        self.QEBin = qebin.copy()
        self.binMask = binMask.copy()
        if self.verbose:
            print("qbin: {}".format(qbin))
            print("ibin: {}".format(ibin))
            print("sebin: {}".format(sebin))
            print("qebin: {}".format(qebin))
            print("binMask: {}".format(binMask))


if __name__ == "__main__":
    # process input arguments
    adict = argparser()
    # transmogrify into kwargs object
    adict = vars(adict)
    # run the reBin program
    reBin(**adict)



# -*- coding: utf-8 -*-
# git.py

import os
import subprocess
import sys


def isRepo(path):
    return os.path.exists(os.path.join(path, ".git"))


def isNBstripoutInstalled():
    out = subprocess.run(
        [sys.executable, "-m", "nbstripout", "--status"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    ).stdout.decode("utf-8")
    return len(out) and "not recognized" not in out


def isNBstripoutActivated():
    out = subprocess.run(
        [sys.executable, "-m", "nbstripout", "--status"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    ).stdout.decode("utf-8")
    return len(out) and "is installed" in out


def checkRepo():
    if not isRepo("."):
        print("Not a GIT repository.")
        return
    # is git installed?
    try:
        import git
    except ImportError:
        print("Could not load git module, is GIT installed and in PATH?")
        return
    # check the repository in detail
    from IPython.display import HTML, display

    repo = git.Repo(".")
    #    currentNB = os.path.basename(currentNBpath())
    try:
        editedOn = repo.git.show(no_patch=True, format="%cd, version %h by %cn", date="iso")
    except git.GitCommandError:
        print("Not a GIT repository.")
        return
    editedOn = editedOn.split(", ")
    opacity = 0.3  # 1.0 if repo.is_dirty() else 0.5
    display(
        HTML(
            '<div style="opacity: {opacity};">'
            "<h3>Document updated on {}</h3>"
            "<h4>({})</h4></div>".format(*editedOn, opacity=opacity)
        )
    )
    if repo.is_dirty():
        edits = repo.git.diff(stat=True)
        import re

        edits = re.sub(r" (\++)", r' <span style="color: green;">\1</span>', edits)
        edits = re.sub(r"(\+)?(-+)(\s)", r'\1<span style="color: red;">\2</span>\3', edits)
        display(
            HTML(
                '<div style="border-style: solid; border-color: darkred; border-width: 1px; '
                'padding: 0em 1em 1em 1em; margin: 1em 0em;">'
                '<h4 style="color: darkred;">There are changes in this repository:</h4>'
                "<pre>"
                + edits
                + "</pre></div>"
            )
        )



# -*- coding: utf-8 -*-
# readdata.py

import tempfile
import warnings
import xml.etree.ElementTree as et
import zipfile
from pathlib import Path

import pandas as pd


def readdata(fpath, q_range=None, read_csv_args=None, print_filename=True):
    """Read a datafile pandas Dataframe
    extract a file_name
    select q-range: q_min <= q <= q_max
    """
    fpath = Path(fpath)
    if print_filename:
        print(f"Reading file '{str(fpath)}'")
    if read_csv_args is None:
        read_csv_args = dict()
    if "sep" not in read_csv_args:
        read_csv_args.update(sep=r"\s+")
    if "names" not in read_csv_args:
        read_csv_args.update(names=("q", "I", "e"))
    if "index_col" not in read_csv_args:
        read_csv_args.update(index_col=False)
    # print("f_read_data, read_csv_args:", read_csv_args) # for debugging

    file_ext = fpath.suffix
    if file_ext.lower() == ".pdh":  # for PDH files
        nrows = pd.read_csv(
            fpath,
            skiprows=2,
            nrows=1,
            usecols=[
                0,
            ],
            sep=r"\s+",
            header=None,
        ).values[0, 0]
        read_csv_args.update(skiprows=5, nrows=nrows)
    df = pd.read_csv(fpath, **read_csv_args)

    # select q-range
    if q_range is not None:
        q_min, q_max = q_range
        df = df[(df.q > q_min) & (df.q < q_max)]

    filename = fpath.stem.split("[")[0]
    return df, filename


readPDH = readdata


def convertValue(val):
    val = val.strip()
    try:
        return int(val)
    except ValueError:
        try:
            return float(val)
        except ValueError:
            pass
    return val


def xmlPDHToDict(root):
    result = {}
    stack = [(root, result)]
    while stack:
        elem, parentCont = stack.pop()
        elemCont = {}
        idx = -1
        key = elem.attrib.pop("key", None)
        if (  # get a unique key, the key can occur in multiple groups in PDH
            key is not None and elem.tag == "group" and elem.attrib.get("id", None) is not None
        ):
            key = elem.attrib.pop("id")
        if (  # skip empty elements with a key only early
            not len(list(elem))
            and not len(elem.attrib)
            and not (elem.text and len(elem.text.strip()))
        ):
            continue
        if elem.tag == "list":
            elemCont = []
        else:  # add attributes & values to dict
            # Attach text, if any
            if elem.text and len(elem.text.strip()):
                if elem.tag in ("value", "reference"):
                    elemCont["value"] = convertValue(elem.text)
                else:
                    elemCont["#text"] = convertValue(elem.text)
            # Attach attributes, if any
            if elem.attrib:
                elemCont.update(
                    {k: convertValue(v) for k, v in elem.attrib.items() if len(v.strip())}
                )
            if key == "unit" and "value" in elemCont:  # fix some units
                elemCont["value"] = elemCont["value"].replace("_", "")
            if "unit" in elemCont:
                elemCont["unit"] = elemCont["unit"].replace("_", "")
            # reduce the extracted dict&attributes
            idx = elemCont.get("index", -1)  # insert last/append if no index given
            value = elemCont.get("value", None)
            if value is not None and (
                len(elemCont) == 1 or (len(elemCont) == 2 and "index" in elemCont)
            ):
                elemCont = value  # contains value only
        parentKey = elem.tag
        if key is not None and parentKey in ("list", "value", "group"):
            # skip one level in hierarchy for these generic containers
            parentKey = key
            key = None
        try:
            if isinstance(parentCont, list):
                parentCont.insert(idx, elemCont)
            elif parentKey not in parentCont:  # add as new list
                if key is None:  # make a list
                    parentCont[parentKey] = elemCont
                else:  # have a key
                    parentCont[parentKey] = {key: elemCont}
            else:  # parentKey exists already
                if not isinstance(parentCont[parentKey], list) and not isinstance(
                    parentCont[parentKey], dict
                ):
                    # if its a plain value before, make a list out of it and append in next step
                    parentCont[parentKey] = [parentCont[parentKey]]
                if isinstance(parentCont[parentKey], list):
                    parentCont[parentKey].append(elemCont)
                elif key is not None:
                    parentCont[parentKey].update({key: elemCont})
                else:  # key is None
                    parentCont[parentKey].update(elemCont)
        except AttributeError:
            raise
        # reversed for correct order
        stack += [(child, elemCont) for child in reversed(list(elem))]
    # fix some entry values, weird Anton Paar PDH format
    try:
        oldts = result["fileinfo"]["parameter"]["DateTime"]["value"]
        # timestamp seems to be based on around 2009-01-01 (a day give or take)
        delta = (39 * 365 + 10) * 24 * 3600
        # make it compatible to datetime.datetime routines
        result["fileinfo"]["parameter"]["DateTime"]["value"] = oldts + delta
    except KeyError:
        pass
    return result


def readPDHmeta(pathPDH):
    """Reads the XML metadata at the end of a .PDH file to a Python dict."""
    pathPDH = Path(pathPDH)
    if pathPDH.suffix.lower() != ".pdh":
        warnings.warn("readPDHmeta() supports .pdh files only!")
        return  # for PDH files
    lines = ""
    with open(pathPDH) as fd:
        lines = fd.readlines()
    nrows = int(lines[2].split()[0])
    xml = "".join(lines[nrows + 5 :])
    return xmlPDHToDict(et.fromstring(xml))


def readSSF(pathSSF):
    """Reads the SAXSquant session file *pathSSF* (.SSF) to a Python dict."""
    pathSSF = Path(pathSSF)
    if pathSSF.suffix.lower() != ".ssf":
        warnings.warn("readSession() supports .ssf files only!")
        return  # for PDH files
    data = ""
    with open(pathSSF, encoding="utf-8-sig") as fd:
        data = fd.read()
    return xmlPDHToDict(et.fromstring(data))


def readSSFZ(pathSSFZ):
    """Extracts and reads the SAXSquant session file (.SSF) to a Python dict.
    The .SSF is embedded in the .SSFZ provided by *pathSSFZ*."""
    assert pathSSFZ.is_file()
    # unpack the SSFZ to a temporary dir
    data = None
    with tempfile.TemporaryDirectory() as tempdir:
        with zipfile.ZipFile(pathSSFZ, "r") as zipfd:
            zipfd.extractall(tempdir)
        # read the session metadata from the extracted SSF file
        pathSSF = next(Path(tempdir).glob("*.ssf"))
        assert pathSSF.is_file()
        data = readSSF(pathSSF)
    return data



# -*- coding: utf-8 -*-
# plotting.py

import matplotlib
import matplotlib.pyplot as plt

from .readdata import readPDH

try:
    # increase the limit for the warning to pop up
    matplotlib.rcParams["figure.max_open_warning"] = 50
except TypeError:  # ignore the error with Sphinx
    pass


def initFigure(fig, width=80, aspectRatio=4.0 / 3.0, quiet=False):
    mmInch = 25.4
    fig.set_size_inches(width / mmInch, width / aspectRatio / mmInch)
    w, h = fig.get_size_inches()
    if not quiet:
        print("initFigure() with ({w:.1f}x{h:.1f}) mm".format(w=w * mmInch, h=h * mmInch))
    return fig


def createFigure(width=80, aspectRatio=4.0 / 3.0, quiet=False, **kwargs):
    """output figure width in mm"""
    fig = plt.figure(
        # tight_layout=dict(pad=0.05),
        **kwargs
    )
    initFigure(fig, width, aspectRatio, quiet)
    return fig


def plotVertBar(ax, xpos, ymax, **kwargs):
    ax.plot((xpos, xpos), (0, ymax), **kwargs)


def plotColor(idx):
    pltcol = plt.rcParams["axes.prop_cycle"].by_key()["color"]
    # print(pltcol)
    pltcol = ["gray", "lightskyblue", "steelblue", "red", "salmon"]
    return pltcol[idx]


def lineWidth():
    return plt.rcParams["lines.linewidth"]


def plotPDH(filename, label, **kwargs):
    """Plot a given .PDH file with the given label (shown in legend) using pandas and readPDH()."""
    q_range = kwargs.pop("q_range", None)
    print_filename = kwargs.pop("print_filename", True)  # default value from readdata()
    df, _ = readPDH(filename, q_range=q_range, print_filename=print_filename)
    df["e"] = df["e"].clip(lower=0)
    defaults = dict(
        yerr="e",
        logx=True,
        logy=True,
        label=label,
        grid=True,
        figsize=(10, 5),
        xlabel=r"$q$ (nm$^{{-1}}$)",
        ylabel="Intensity",
        ecolor="lightgray",
    )
    for k, v in defaults.items():
        if k not in kwargs:
            kwargs[k] = v
    df.plot("q", "I", **kwargs)



# -*- coding: utf-8 -*-
# tests/readdata.py

import json
import os
import tempfile
import zipfile
from pathlib import Path

import numpy

from jupyter_analysis_tools import readdata, readPDHmeta, readSSF, readSSFZ

pathPDH1 = Path("testdata/S2842 water.pdh")
pathPDH2 = Path("testdata/S2843[9].pdh")  # desmeared silica measurement
pathSSFZ = Path("testdata/2015-03-20-Silica.ssfz")


def test_readdata1(capsys):
    assert pathPDH1.is_file()
    df, fn = readdata(pathPDH1)
    captured = capsys.readouterr()
    assert captured.out == f"Reading file 'testdata{os.sep}S2842 water.pdh'\n"
    assert fn == "S2842 water"
    assert df.shape == (1280, 3)
    assert df.columns.tolist() == ["q", "I", "e"]
    assert df.dtypes.tolist() == [numpy.float64, numpy.float64, numpy.float64]
    # checking the first data values
    assert numpy.all(
        df.loc[:2].values
        == numpy.array(
            [
                [-1.005583e00, 5.555556e-08, 2.754402e-08],
                [-9.989474e-01, 3.611111e-07, 6.568830e-08],
                [-9.923112e-01, 3.055556e-07, 6.120415e-08],
            ]
        )
    )
    # and checking the last data values
    assert numpy.all(
        df.loc[df.shape[0] - 3 :].values
        == numpy.array(
            [
                [7.381979e00, 2.972222e-06, 1.792166e-07],
                [7.388376e00, 2.944444e-06, 1.436040e-07],
                [7.394774e00, 2.388889e-06, 1.548690e-07],
            ]
        )
    )
    assert numpy.all(df.median().values == numpy.array([3.233221e00, 5.826389e-05, 8.835466e-07]))


def test_readdata2(capsys):
    # test another file with different naming scheme
    assert pathPDH2.is_file()
    df, fn = readdata(pathPDH2, print_filename=False)
    captured = capsys.readouterr()
    assert captured.out == ""
    assert fn == "S2843"
    assert df.shape == (427, 3)
    assert df.columns.tolist() == ["q", "I", "e"]
    assert numpy.all(df.median().values == numpy.array([1.470428, 0.01907878, 0.01353293]))


def test_readPDHmeta1():
    # read the meta data of the PDH file
    assert pathPDH1.is_file()
    data = readPDHmeta(pathPDH1)
    assert data is not None
    assert isinstance(data, dict)
    assert len(data)  # there should be data at all

    # writing the test JSON for comparisons on updates
    # with open(pathPDH1.with_suffix(".json"), "w") as fd:
    #     json.dump(data, fd, indent=4)

    # write the JSON formatted metadata to disk, read it back in and compare
    # it with the expected reference from testdata dir
    with open(pathPDH1.with_suffix(".json")) as fdRef, tempfile.TemporaryFile("w+") as fdNew:
        json.dump(data, fdNew, indent=4)
        fdNew.seek(0)
        assert fdRef.read() == fdNew.read()

    # if there are changes, use this to investigate:
    # from difflib import Differ
    # with open('file1.txt') as f1, open('file2.txt') as f2:
    #    differ = Differ()
    #    for line in differ.compare(f1.readlines(), f2.readlines()):
    #        print(line) # ignore lines starting with ' ' (no change)


def test_readSSF():
    assert pathSSFZ.is_file()
    # unpack the SSFZ to a temporary dir
    with tempfile.TemporaryDirectory() as tempdir:
        with zipfile.ZipFile(pathSSFZ, "r") as zipfd:
            zipfd.extractall(tempdir)
        # read the session meta data from the extracted SSF file
        pathSSF = next(Path(tempdir).glob("*.ssf"))
        assert pathSSF.is_file()
        data = readSSF(pathSSF)
        assert data is not None
        assert isinstance(data, dict)
        assert len(data)  # there should be data at all
        # writing the test JSON for comparisons on updates
        # with open(pathSSFZ.with_suffix(".ssf.json"), "w") as fd:
        #    json.dump(data, fd, indent=4)
        # write the JSON formatted session data to disk
        # and compare it with the expected JSON file from testdata dir
        with open(pathSSFZ.with_suffix(".ssf.json")) as fdRef, tempfile.TemporaryFile(
            "w+"
        ) as fdNew:
            json.dump(data, fdNew, indent=4)
            fdNew.seek(0)
            assert fdRef.read() == fdNew.read()


def test_readSSFZ():
    assert pathSSFZ.is_file()
    data = readSSFZ(pathSSFZ)
    # compare with the expected JSON file from testdata dir
    with open(pathSSFZ.with_suffix(".ssf.json")) as fdRef:
        dataRef = json.load(fdRef)
        assert dataRef == data



# -*- coding: utf-8 -*-
# analysis.py

import numpy as np


# from https://stackoverflow.com/a/22357811
# and https://github.com/joferkington/oost_paper_code/blob/master/utilities.py#L167
# (code with MIT License)
def getModZScore(points):
    """
    Returns a boolean array with True if points are outliers and False
    otherwise.
    **Note**:
    Similar to https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.zscore.html
    but uses the median instead of the mean.

    :param points: An numobservations by numdimensions array of observations
    :param thresh: The modified z-score to use as a threshold. Observations with
        a modified z-score (based on the median absolute deviation) greater
        than this value will be classified as outliers.

    Returns
    -------
    mask: numpy array
        A numobservations-length boolean array.

    References
    ----------
    Boris Iglewicz and David Hoaglin (1993), "Volume 16: How to Detect and
    Handle Outliers", The ASQC Basic References in Quality Control:
    Statistical Techniques, Edward F. Mykytka, Ph.D., Editor.
    """
    if len(points.shape) == 1:
        points = points[:, None]
    median = np.median(points, axis=0)
    diff = np.sqrt(np.sum((points - median) ** 2, axis=-1))
    med_abs_deviation = np.median(diff)

    # scale being the inverse of the standard normal quantile function at 0.75,
    # which is approximately 0.67449, see also:
    # https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.median_abs_deviation.html
    # modified_z_score = 0.6745 * diff / med_abs_deviation
    # let this indicator be =1 for the same data, makes it more intuitive to understand
    modified_z_score = diff / med_abs_deviation

    return modified_z_score



# -*- coding: utf-8 -*-
# __init__.py

__version__ = "1.4.3"

from .binning import reBin
from .git import checkRepo, isNBstripoutActivated, isNBstripoutInstalled, isRepo
from .plotting import createFigure, plotPDH
from .readdata import readdata, readPDH, readPDHmeta, readSSF, readSSFZ
from .utils import naturalKey, setLocaleUTF8
from .widgets import PathSelector, showBoolStatus

setLocaleUTF8()



# -*- coding: utf-8 -*-
# utils.py

import contextlib
import copy
import itertools
import locale
import os
import platform
import re
import subprocess
import sys
from pathlib import Path

import numpy as np

indent = "    "


def setLocaleUTF8():
    """Fix the Jupyter locale which is not UTF-8 by default on Windows."""
    locOld = locale.getpreferredencoding(False).lower()

    def getpreferredencoding(do_setlocale=True):
        return "utf-8"

    locale.getpreferredencoding = getpreferredencoding
    locNew = locale.getpreferredencoding(False)
    if locOld != locNew:
        print(f"Updated locale from {locOld} -> {locNew}.")


def isLinux():
    return platform.system().lower() in "linux"


def isMac():
    return platform.system().lower() in "darwin"


def isWindows():
    return platform.system().lower() in "windows"


def isList(obj):
    """Return true if the provided object is list-like including a numpy array but not a string.

    >>> isList([1, 2, 'a'])
    True
    >>> isList(tuple((1, 2, 'a')))
    True
    >>> import numpy
    >>> isList(numpy.arange(5))
    True
    >>> isList("dummy")
    False
    >>> isList(None)
    False
    """
    return isinstance(obj, (list, tuple, np.ndarray))


def shortenWinPath(path):
    if not isWindows():
        return path
    import win32api

    return win32api.GetShortPathName(path)


def appendToPATH(parentPath, subdirs=None, verbose=False):
    """Adds the given path with each subdirectory to the PATH environment variable."""
    parentPath = Path(parentPath)
    if not parentPath.is_dir():
        return  # nothing to do
    if subdirs is None:
        subdirs = ["."]
    sep = ";" if isWindows() else ":"
    PATH = os.environ["PATH"].split(sep)
    for path in subdirs:
        path = parentPath / path
        if verbose:
            print(indent, path, "[exists: {}]".format(path.is_dir()))
        if path not in PATH:
            PATH.append(str(path))
    os.environ["PATH"] = sep.join(PATH)


def addEnvScriptsToPATH():
    """Prepends the *Scripts* directory of the current Python environment base directory to systems
    PATH variable.

    It is intended for Conda (Miniforge) environments on Windows that do not have this in their PATH
    environment variable, causing them to miss many commands provided from this location.
    """
    envPath = [p for p in sys.path if p.endswith("Lib")]
    if not envPath:
        return  # probably not a Miniforge environment
    envPath = envPath[0]
    envPath = Path(envPath).parent / "Scripts"
    sep = ";" if isWindows() else ":"
    environPATH = os.environ["PATH"].split(sep)
    # print(environPATH)
    if envPath.exists() and str(envPath) not in environPATH:
        environPATH = [str(envPath)] + environPATH
        os.environ["PATH"] = sep.join(environPATH)


def networkdriveMapping(cmdOutput: str = None, resolveNames: bool = True):
    """Returns a dict of mapping drive letters to network paths (on Windows)."""
    if isWindows():
        if cmdOutput is None:
            proc = subprocess.run(["net", "use"], capture_output=True, text=True, encoding="cp850")
            cmdOutput = proc.stdout

        def resolveFQDN(uncPath):
            if not resolveNames:
                return uncPath
            parts = uncPath.split("\\")
            idx = [i for i, part in enumerate(parts) if len(part)][0]
            proc = subprocess.run(
                ["nslookup", parts[idx]], capture_output=True, text=True, encoding="cp850"
            )
            res = [line.split() for line in proc.stdout.splitlines() if line.startswith("Name:")]
            if len(res) and len(res[0]) == 2:
                parts[idx] = res[0][1]
            return "\\".join(parts)

        rows = [line.split() for line in cmdOutput.splitlines() if "Windows Network" in line]
        rows = {
            row[1]: resolveFQDN(row[2])
            for row in rows
            if row[1].endswith(":") and row[2].startswith(r"\\")
        }
        return rows
    else:  # Linux (tested) or macOS (untested)
        if cmdOutput is None:
            proc = subprocess.run(["mount"], capture_output=True, text=True)
            cmdOutput = proc.stdout

        def parse(line):
            # position of last opening parenthesis, start of options list
            lastParen = list(i for i, c in enumerate(line) if "(" == c)[-1]
            line = line[:lastParen].strip()
            spaces = list(i for i, c in enumerate(line) if " " == c)
            fstype = line[spaces[-1] :].strip()  # last remaining word is the filesystem type
            line = line[: spaces[-2]].strip()  # strip the 'type' indicator as well
            sepIdx = line.find(" on /")  # separates destination from mount point
            dest = line[:sepIdx].strip()
            mountpoint = line[sepIdx + 4 :].strip()
            yield (mountpoint, dest, fstype)

        return {
            mp: dst
            for line in cmdOutput.strip().splitlines()
            for (mp, dst, fstype) in parse(line)
            if fstype in ("nfs", "cifs", "sshfs", "afs", "ext4")
        }
    return {}


def makeNetworkdriveAbsolute(filepath, cmdOutput: str = None, resolveNames: bool = True):
    """Replaces the drive letter of the given path by the respective network path, if possible."""
    if filepath.drive.startswith(r"\\"):
        return filepath  # it's a UNC path already
    if isWindows():
        drivemap = networkdriveMapping(cmdOutput=cmdOutput, resolveNames=resolveNames)
        prefix = drivemap.get(filepath.drive, None)
        if prefix is not None:
            filepath = Path(prefix).joinpath(*filepath.parts[1:])
    else:  # Linux or macOS
        drivemap = networkdriveMapping(cmdOutput=cmdOutput, resolveNames=resolveNames)
        # search for the mountpoint, starting with the longest, most specific, first
        for mp, target in sorted(drivemap.items(), key=lambda tup: len(tup[0]), reverse=True):
            if filepath.is_relative_to(mp):
                return Path(target).joinpath(filepath.relative_to(mp))
    return filepath


def checkWinFor7z():
    """Extend the PATH environment variable for access to the 7-zip executable."""
    if not isWindows():
        return  # tests below are intended for Windows
    sevenzippath = r"C:\Program Files\7-Zip"
    if not os.path.isdir(sevenzippath):
        print(
            "7-Zip not found in '{}'.\n".format(sevenzippath)
            + "7-Zip is required for managing data files and results!."
        )
        return
    print("Adding the following directory to $PATH:")
    appendToPATH(sevenzippath)
    print("\nUpdated PATH:")
    for path in os.environ["PATH"].split(";"):
        print(indent, path)


def extract7z(fn, workdir=None):
    assert os.path.isfile(os.path.join(workdir, fn)), "Provided 7z archive '{}' not found!".format(
        fn
    )
    print(f"Extracting '{fn}': ")
    proc = subprocess.run(
        ["7z", "x", fn],
        cwd=workdir,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    print(proc.stdout.decode(errors="ignore"))
    if len(proc.stderr):
        print("## stderr:\n", proc.stderr.decode(errors="ignore"))


# https://stackoverflow.com/a/13847807
@contextlib.contextmanager
def pushd(new_dir):
    previous_dir = os.getcwd()
    os.chdir(new_dir)
    yield
    os.chdir(previous_dir)


def setPackage(globalsdict):
    """Sets the current directory of the notebook as python package to make relative module imports
    work.

    Usage: `setPackage(globals())`
    """
    path = Path().resolve()
    searchpath = str(path.parent)
    if searchpath not in sys.path:
        sys.path.insert(0, searchpath)
    globalsdict["__package__"] = path.name
    globalsdict["__name__"] = path.name
    print(f"Setting the current directory as package '{path.name}': \n  {path}.")


def grouper(iterable, n, fillvalue=None):
    """Returns an iterator over a list of tuples (grouping) for a given flat iterable."""
    args = [iter(iterable)] * n
    return itertools.zip_longest(*args, fillvalue=fillvalue)


def fmtErr(val, std, precision=2, width=None):
    """Formats a given value and its stdandard deviation to physics notation, e.g. '1.23(4)'."""
    if width is None:
        width = ""
    fmt = "{:" + str(width) + "." + str(precision) + "f}({:.0f})"
    # print("fmtErr val:", val, "std:", std)
    return fmt.format(val, std * 10 ** (precision))


def updatedDict(d, key, value):
    """Implements the \\|= operator for dict in Python version <3.9."""
    dd = copy.copy(d)
    dd[key] = value
    return dd


def naturalKey(name):
    """Split string into list of strings and integers. Use as *key* function for sorting files."""
    return [int(text) if text.isdigit() else text.lower() for text in re.split(r"(\d+)", name)]



# -*- coding: utf-8 -*-
# distrib.py

import itertools
from collections.abc import Iterable
from operator import itemgetter

import matplotlib.font_manager as font_manager
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy.integrate
import scipy.interpolate

from .plotting import plotVertBar
from .utils import grouper


def integrate(xvec, yvec):
    try:
        return abs(scipy.integrate.simps(yvec, x=xvec))
    except AttributeError:
        return abs(scipy.integrate.simpson(yvec, x=xvec))


def normalizeDistrib(x, y, u=None):
    x = x.values if isinstance(x, pd.Series) else x
    y = y.values if isinstance(y, pd.Series) else y
    # normalize the distribution to area of 1
    norm = integrate(x, y)
    # print("CONTINs norm", norm)
    y /= norm
    if u is not None:
        u /= norm
    return x, y, u


def area(xvec, yvec, showArea=True):
    """Returns a string with the area value of the given discrete curve points."""
    return r" $\int${:.3g}".format(integrate(xvec, yvec)) if showArea else ""


def findPeakRanges(x, y, tol=1e-16):
    """Returns the location of data/peak above a base line.
    Assumes it touches the baseline before and after. For distributions.
    *tol*: Multiplied by Y to produce a threshold to distinguish noise/artifacts from peaks."""
    x = x.values if isinstance(x, pd.Series) else x
    y = y.values if isinstance(y, pd.Series) else y
    # look at all data above zero, get their array indices
    indices = np.where(y > tol * y.max())[0]
    # segmentation: look where continous groups of indices start and end
    indexGroups = np.where(np.diff(indices) > 1)[0]
    ranges = []
    istart = indices[0]

    def appendPeakRange(start, end):
        # print("appending", start, end, end-start)
        start, end = max(start - 1, 0), min(end + 1, len(x) - 1)
        monotony = np.sign(np.diff(y[start : end + 1]))
        if not all(monotony == monotony[0]):
            # avoid monotonously increasing/decreasing peaks -> unwanted artefacts
            ranges.append((start, end))

    for idx in indexGroups:
        appendPeakRange(istart, indices[idx])  # add the new range to the list
        istart = indices[idx + 1]  # start new range
    appendPeakRange(istart, indices[-1])
    # print("findPeakRanges", ranges)
    return ranges


def findLocalMinima(peakRanges, xarr, yarr, doPlot=False, verbose=False):
    """Identify local (non-zero) minima within given peak ranges and separate those
    bimodal ranges into monomodal ranges, thus splitting up the peak range if it contains
    maxima connected by non-zero minima. Returns a list of index tuples indicating the
    start and end of each peak. Uses 4th order spline fitting and its derivative
    for finding positions of local minima."""
    # print("findLocalMinima", peakRanges)
    newRanges = []
    if doPlot:
        plt.figure(figsize=(15, 5))
    for ip, (istart, iend) in enumerate(peakRanges):
        if verbose:
            print((istart, iend), xarr[istart], xarr[iend])
        if iend - istart < 5:  # skip this, can't be fitted and no sub-peaks are likely
            newRanges.append((istart, iend))
            continue
        while yarr[istart] <= 0.0 and istart < iend:
            istart += 1  # exclude leading zero
        while yarr[iend] <= 0.0 and istart < iend:
            iend -= 1  # exclude trailing zero
        if istart == iend:
            continue
        if verbose:
            print((istart, iend))
        x, y = xarr[istart : iend + 1], yarr[istart : iend + 1]
        try:
            spline = scipy.interpolate.InterpolatedUnivariateSpline(x, y, k=4)
        except Exception:
            print(f"Warning: Could not findLocalMinima() within {(istart, iend)}!")
            newRanges.append((istart, iend))
            continue
        # if verbose: print(spline(x))
        deriv = spline.derivative()
        # if verbose: print(deriv(x))
        roots = deriv.roots()
        # get indices of roots and ignore any duplicate indices
        rootIdx = set(np.argmin(np.abs(xarr[:, np.newaxis] - roots[np.newaxis, :]), axis=0))
        rootIdx.add(istart)
        rootIdx.add(iend)
        rootIdx = sorted(rootIdx)
        # if rootIdx[0] == istart: # omit the first root at the beginning
        #    rootIdx = rootIdx[1:]
        if verbose:
            print((istart, iend), len(roots), roots, rootIdx)
        if doPlot:
            plt.subplot(1, len(peakRanges), ip + 1)
            radGrid = np.linspace(x[0], x[-1], 200)
            plt.plot(x, y, label="data")
            plt.plot(radGrid, spline(radGrid), label="spline"),
            plt.ylabel("data & spline approx.")
            handles1, labels1 = plt.gca().get_legend_handles_labels()
            [
                plotVertBar(plt, xarr[i], spline(radGrid).max(), color="blue", ls=":")
                for i in rootIdx
            ]
            plt.gca().twinx()
            plt.plot(radGrid, deriv(radGrid), label="deriv. spline", color="green")
            plt.ylabel("1st derivative")
            handles2, labels2 = plt.gca().get_legend_handles_labels()
            plt.grid()
            plt.legend(handles1 + handles2, labels1 + labels2)
        peakBoundaries = rootIdx[::2]
        if verbose:
            print(peakBoundaries)
        newRanges += [tuple(peakBoundaries[i : i + 2]) for i in range(len(peakBoundaries) - 1)]
    if verbose:
        print(newRanges)
    return newRanges


def getLargestPeaks(peakRanges, xarr, yarr, count=1):
    def peakRangeArea(peakRange):
        return integrate(
            xarr[peakRange[0] : peakRange[1] + 1], yarr[peakRange[0] : peakRange[1] + 1]
        )

    return sorted(peakRanges, key=peakRangeArea, reverse=True)[:count]


class Moments(dict):
    @staticmethod
    def nthMoment(x, weights, n):
        """Calculates the nth moment of the given distribution weights."""
        center = 0
        if n > 0:  # calculate the mean first
            center = np.average(x, weights=weights) if sum(weights) else 0.0
            #          np.sqrt(u**2)/len(u) # center uncertainty
        if n == 1:
            return center  # the mean
        var = 1.0
        if n > 1:
            var = np.sum(weights * (x - center) ** 2) / np.sum(weights)
        if n == 2:
            return var  # the variance
        return np.sum(weights * (x - center) ** n) / np.sum(weights) / var**n

    @classmethod
    def fromData(cls, x, y):
        store = cls()
        mean, var, skew, kurt = [cls.nthMoment(x, y, i) for i in range(1, 5)]
        store["area"] = integrate(x, y)
        store["mean"] = mean
        store["var"] = var
        store["skew"] = skew
        store["kurt"] = kurt
        return store

    @property
    def area(self):
        return self["area"]

    @property
    def mean(self):
        return self["mean"]

    @property
    def var(self):
        return self["var"]

    @property
    def skew(self):
        return self["skew"]

    @property
    def kurt(self):
        return self["kurt"]

    def __str__(self):
        return "\n".join(
            [
                "{: <4s}: {: 9.2g}".format(k, self[k])
                for k in ("area", "mean", "var", "skew", "kurt")
            ]
        )

    @staticmethod
    def logNormParFromMoments(mean, var, N=1.0):
        # SASfit manual, 6.4. Log-Normal distribution
        median = mean**2 / np.sqrt(var + mean**2)
        sigma = np.sqrt(np.log(mean**2 / median**2))
        return {"N": N, "sigma": sigma, "median": median}

    def logNormPar(self, N=1.0):
        return self.logNormParFromMoments(self.mean, self.var, N=N)


class Distribution:
    x, y, u = None, None, None
    peaks = None  # list of peak (start, end) indices pointing into x,y,u
    color = None
    xlabel = None
    plotAxes, plotAxisIdx = None, 0

    def __init__(self, xvec, yvec, uvec, xlabel=None, maxPeakCount=None):
        self.xlabel = getattr(xvec, "name", None)
        xvec = xvec.values if isinstance(xvec, pd.Series) else xvec
        yvec = yvec.values if isinstance(yvec, pd.Series) else yvec
        uvec = uvec.values if isinstance(uvec, pd.Series) else uvec
        self.x, self.y, self.u = normalizeDistrib(xvec, yvec, uvec)
        if xlabel is not None:
            self.xlabel = xlabel
        self.peaks = findPeakRanges(self.x, self.y, tol=1e-6)
        # refine the peak ranges containing multiple maxima
        self.peaks = findLocalMinima(self.peaks, self.x, self.y)
        # For a given list of peaks (by start/end indices) return only those
        # whose ratio of amount to uncertainty ratio is always below the given max. ratio
        # maxRatio = 1.5
        # self.peakRanges = [(istart, iend) for istart, iend in self.peakRanges
        #                    if maxRatio > 1/np.median(self.y[istart:iend+1]/self.u[istart:iend+1])]
        # Sort the peaks by area and use the largest (last) only, assuming monomodal distributions
        if maxPeakCount:
            self.peaks = getLargestPeaks(self.peaks, self.x, self.y, count=maxPeakCount)

    def peakData(self, peakRange):
        return (
            self.x[peakRange[0] : peakRange[1] + 1],
            self.y[peakRange[0] : peakRange[1] + 1],
            self.u[peakRange[0] : peakRange[1] + 1],
        )

    def uncertRatioMedian(self, peakRange):
        _, y, u = self.peakData(peakRange)
        return 1.0 / np.median(y / u)

    @staticmethod
    def getBarWidth(xvec):
        return np.concatenate((np.diff(xvec)[:1], np.diff(xvec)))

    def plotPeak(self, peakRange, mom, momLo, momHi, dp, dpLo, dpHi, showFullRange=False, ax=None):
        """
        *showFullRange*: Set the x range to cover the whole distribution instead of the peak only.
        """
        x, y, u = self.peakData(peakRange)
        if not ax:
            ax = plt.gca()
        # ax.plot(x, y, 'o', color=cls.color)
        lbl, fmt = [], "{: <7s} {: 9.2g} ±{: 9.2g}"
        for k in "area", "median", "var", "skew", "kurt":
            if k == "median":
                lbl.append(
                    fmt.format(
                        "median:",
                        dp["median"],
                        max(abs(dp["median"] - dpLo["median"]), abs(dpHi["median"] - dp["median"])),
                    )
                )
            else:
                lbl.append(
                    fmt.format(k + ":", mom[k], max(abs(mom[k] - momLo[k]), abs(momHi[k] - mom[k])))
                )
        lbl.append("LogNorm: " + distrParToText(dp)[0])
        ax.bar(x, y, width=self.getBarWidth(x), color=self.color, alpha=0.5, label="\n".join(lbl))
        ax.fill_between(
            x,
            np.maximum(0, y - u),
            y + u,
            color="red",
            lw=0,
            alpha=0.1,
            label=f"uncertainties (lvl: {self.uncertRatioMedian(peakRange):.3g})",
        )
        if showFullRange:
            ax.set_xlim((self.x.min(), self.x.max()))
        ax.set_xlabel(self.xlabel)
        ax.grid(True)
        legend = ax.legend(prop=font_manager.FontProperties(family="monospace"))
        # make the legend background more transparent
        legend.get_frame().set_alpha(None)
        legend.get_frame().set_facecolor((1, 1, 1, 0.2))

    def plot(self, ax, distPar, name=""):
        """plot complete distribution as loaded from file"""
        lbl = (
            "from file, "
            + name
            + area(self.x, self.y, showArea=True)
            + "\n"
            + distrParLatex(distPar[0])
        )
        ax.fill_between(
            self.x,
            self.y,
            # width=GenericResult.getBarWidth(self.x),
            color=self.color,
            alpha=0.5,
            label=lbl,
        )
        # ax.errorbar(self.x, self.y, yerr=self.u, lw=lineWidth()*2, label=lbl)
        ax.fill_between(
            self.x,
            np.maximum(0, self.y - self.u),
            self.y + self.u,
            color="red",
            lw=0,
            alpha=0.1,
            label="uncertainties",
        )
        ax.set_xlabel(self.xlabel)
        ax.legend()
        ax.grid()
        ax.set_xscale("log")

    def moments(self):
        def momentsByPeak():
            for peakRange in self.peaks:
                x, y, u = self.peakData(peakRange)
                N = integrate(x, y)
                mom = Moments.fromData(x, y)
                momLo = Moments.fromData(x, np.maximum(0, y - u))
                momHi = Moments.fromData(x, y + u)
                lnp = mom.logNormPar(N=N)
                lnpLo = momLo.logNormPar(N=N)
                lnpHi = momHi.logNormPar(N=N)
                yield (peakRange, mom, momLo, momHi, lnp, lnpLo, lnpHi)

        # return a dict of lists, addressable by peak index
        return dict(
            zip(
                ["peakRange", "mom", "momLo", "momHi", "lnp", "lnpLo", "lnpHi"],
                zip(*[m for m in momentsByPeak()]),
            )
        )

    def peakDistrPar(self, plotAxes=None, plotAxisStart=0, **plotPeakKwargs):
        momentsAndLogNormPar = self.moments()
        if plotAxes is not None:
            for i, peakRange in enumerate(momentsAndLogNormPar["peakRange"]):
                plotPeakKwargs["ax"] = plotAxes[plotAxisStart + i]
                self.plotPeak(*[v[i] for v in momentsAndLogNormPar.values()], **plotPeakKwargs)
        return momentsAndLogNormPar["lnp"], momentsAndLogNormPar["mom"]


def distrParToText(logNormPar):
    """
    >>> distrParToText({'N':1.1, 'sigma':0.15, 'median':33.1234e-9})
    ['median=3.3e-08 sigma=0.15 N=1.1']
    >>> distrParToText({'N':1.2e13, 'sigma':0.15, 'median':3.1234})
    ['median=3.1 sigma=0.15 N=1.2e+13']
    >>> distrParToText({'N':(1.,2.), 'sigma':(.2,.4), 'median':(40e-9,7e-8)})
    ['median_0=4e-08 sigma_0=0.20 N_0=1', 'median_1=7e-08 sigma_1=0.40 N_1=2']
    """
    fmt = {"median": "{:.2g}", "sigma": "{:.2f}", "N": "{:.2g}"}
    order = {key: list(fmt.keys()).index(key) for key in fmt.keys()}
    return [
        " ".join(p)
        for p in grouper(
            list(
                zip(
                    *sorted(
                        [
                            (
                                i * 10 + order[key],
                                f"{key}"
                                + (f"_{i}" if isinstance(vals, Iterable) else "")
                                + "="
                                + fmt[key].format(v),
                            )
                            for key, vals in logNormPar.items()
                            for i, v in enumerate(vals if isinstance(vals, Iterable) else [vals])
                        ],
                        key=itemgetter(0),
                    )
                )
            )[-1],
            3,
        )
    ]


def distrParLatex(distrPar, *kwargs):
    r"""
    >>> distrParLatex({'N':1.1, 'sigma':0.15, 'median':33e-9})
    '$median=3.3e-08\\;sigma=0.15\\;N=1.1$'
    >>> distrParLatex({'N':(1.,2.), 'sigma':(.2,.4), 'median':(40e-9,7e-8)})
    '$median_0=4e-08\\;sigma_0=0.20\\;N_0=1$\n$median_1=7e-08\\;sigma_1=0.40\\;N_1=2$'
    """
    return "\n".join(["$" + txt.replace(" ", r"\;") + "$" for txt in distrParToText(distrPar)])


def distrParToFilename(distrPar, prefix=""):
    """
    >>> distrParToFilename({'N':1.1, 'sigma':0.15, 'median':33e-9})
    '_median=3.3e-08_sigma=0.15_N=1.1'
    >>> distrParToFilename({'N':(1.,2.), 'sigma':(.2,.4), 'median':(40e-9,7e-8)})
    '_median_0=4e-08_sigma_0=0.20_N_0=1_median_1=7e-08_sigma_1=0.40_N_1=2'
    """
    return "_".join([prefix] + distrParToText(distrPar)).replace(" ", "_")


def distrParFromFilename(fn):
    """
    >>> distrParFromFilename('_median=_33_sigma=0.15_N=1.1') == {'N':1.1, 'sigma':0.15, 'median':33}
    True
    >>> fn = '_median_0=4e-08_sigma_0=0.20_N_0=1_median_1=7e-08_sigma_1=0.40_N_1=2'
    >>> distrParFromFilename(fn) == {'N':(1.,2.), 'sigma':(.2,.4), 'median':(40e-9,7e-8)}
    True
    """
    fn = fn.split("=")
    fn = [elem.lstrip("_") for elem in fn]
    fn = [(elem.split("_", maxsplit=1) if elem[0].isnumeric() else [elem]) for elem in fn]
    fn = list(itertools.chain(*fn))
    result = {}
    for k, v in grouper(fn, 2):
        key = k.split("_")[0]
        value = float(v) if "median" == key else float(v)
        result[key] = (
            value
            if key not in result
            else (
                result[key] + (value,) if isinstance(result[key], tuple) else (result[key], value)
            )
        )
    return result
